"""

"""
# Full imports
import json

# Partial imports
from abc import ABC, abstractmethod
from genson import SchemaBuilder
from jsonschema import Draft7Validator
from pytups import SuperDict
from typing import List
from datetime import datetime, timedelta

# Imports from internal modules
from .read_tools import read_excel, is_xl_type


class InstanceSolutionCore(ABC):
    """
    Common interface for the instance and solution templates
    """

    def __init__(self, data: dict):
        self.data = SuperDict.from_dict(data)

    @property
    def data(self) -> dict:
        """
        input data (not necessarily in json-format)
        """
        return self._data

    @data.setter
    def data(self, value: dict):
        self._data = value

    def _get_property(self, key, prop) -> SuperDict:
        return self.data[key].get_property(prop)

    @classmethod
    def from_dict(cls, data: dict) -> "InstanceSolutionCore":
        """
        :param data: json-schema in a dictionary format

        :return: an object initialized from the dict json-schema
        """
        return cls(data)

    def to_dict(self) -> dict:
        """
        :return: a dictionary with the json-schema representation
        """
        return self.data

    @classmethod
    def from_json(cls, path: str) -> "InstanceSolutionCore":
        """
        :param path: path to json-schema json file

        :return: an object initialized from the json-schema formatted json file
        """
        with open(path, "r") as f:
            data_json = json.load(f)
        return cls.from_dict(data_json)

    def to_json(self, path: str) -> None:
        """
        :param path: path to json-schema json file

        writes a json file with the json-schema representation of the object
        """

        data = self.to_dict()
        with open(path, "w") as f:
            json.dump(data, f, indent=4, sort_keys=True)

    @property
    @abstractmethod
    def schema(self) -> dict:
        """
        a dictionary representation of the json-schema for the object
        """

        raise NotImplementedError()

    def check_schema(self) -> List:
        """
        checks that the json-schema export complies with the defined schema

        :return: a list of errors
        """

        validator = Draft7Validator(self.schema)
        data = self.to_dict()
        if not validator.is_valid(data):
            return [e for e in validator.iter_errors(data)]
        return []

    def generate_schema(self) -> dict:
        """
        :return: a dict json-schema based on the current data
        """
        builder = SchemaBuilder()
        builder.add_schema({"type": "object", "properties": {}})
        builder.add_object(self.to_dict())
        return builder.to_schema()

    @classmethod
    def from_excel(cls, path: str) -> "InstanceSolutionCore":
        """
        Read an entire excel file.

        :param path: path of the excel file
        :return: a dict with a list of dict (records format) for each table.
        """

        param_tables_names = cls.__get_parameter_tables_names()

        tables = read_excel(path, param_tables_names)
        return cls.from_dict(tables)

    @classmethod
    def __get_parameter_tables_names(cls) -> List:
        """
        :return: The names of the table of the schema that are parameter tables
        """
        json_schema = cls(dict()).schema
        if json_schema.get("properties", None):
            return [
                table_name
                for table_name, content in json_schema["properties"].items()
                if content["type"] == "object"
            ]
        return []

    def to_excel(self, path: str):
        """
        Write data to excel.

        :param path: path or name of the excel file
        :return: nothing
        """
        try:
            import pandas as pd
        except (ImportError, ModuleNotFoundError):
            raise Exception("You must install pandas package to use this method")

        is_xl_type(path)

        data = self.to_dict()

        with pd.ExcelWriter(path) as writer:
            for table in data.keys():
                content = data[table]
                if isinstance(content, list):
                    pd.DataFrame.from_records(content).to_excel(
                        writer, table, index=False
                    )
                elif isinstance(content, dict):
                    pd.DataFrame.from_dict(content, orient="index").to_excel(
                        writer, table, header=False
                    )

    """
    DATA HANDLING METHODS
    """

    @staticmethod
    def dict_to_int_or_float(data_dict):
        """
        Tranforms a dictionary to change all strings into integer of floating numbers if the strings
            represent numbers
        For example: Transforms {a: '4', b: {c: '7', d: ['8.7', '9']}}
            into {a: 4, b: {c: 7, d: [8.7, 9]}}
        """
        for key in data_dict.keys():
            if isinstance(data_dict[key], str):
                if data_dict[key].isnumeric():
                    data_dict[key] = int(data_dict[key])
                else:
                    try:
                        fl = float(data_dict[key])
                        data_dict[key] = fl
                    except ValueError:
                        pass
            elif isinstance(data_dict[key], list):
                if isinstance(data_dict[key][0], str):
                    try:
                        data_dict[key] = list(map(int, data_dict[key]))
                    except ValueError:
                        try:
                            data_dict[key] = list(map(float, data_dict[key]))
                        except ValueError:
                            pass
                elif isinstance(data_dict[key][0], dict):
                    data_dict[key] = list(
                        map(InstanceSolutionCore.dict_to_int_or_float, data_dict[key])
                    )
            elif isinstance(data_dict[key], dict):
                data_dict[key] = InstanceSolutionCore.dict_to_int_or_float(
                    data_dict[key]
                )
        return dict(data_dict)

    @staticmethod
    def from_element_or_list_to_dict(element_or_list):
        """
        Converts a list into a dictionary indexed by the field 'index' of each element of the list.
        If the input is not a list, it is converted into a list before converting to a dictionary
        For example: [{'index': 4, 'value': 5}, {'index': 7, 'value': 8}]
            is transformed to {4: {'index': 4, 'value': 5}, 7: {'index': 7, 'value': 8}}
        """
        if not isinstance(element_or_list, list):
            element_or_list = [element_or_list]
        return {int(el["index"]): el for el in element_or_list}

    @staticmethod
    def get_date_from_string(string: str) -> datetime:
        """
        Returns a datetime object from an hour-string in format 'YYYY-MM-DD'
        """
        return datetime.strptime(string, "%Y-%m-%d")

    @staticmethod
    def get_datetime_from_string(string: str) -> datetime:
        """
        Returns a datetime object from an hour-string in format 'YYYY-MM-DDTh:m'
        """
        return datetime.strptime(string, "%Y-%m-%dT%H:%M")

    @staticmethod
    def get_datetimesec_from_string(string: str) -> datetime:
        """
        Returns a datetime object from an hour-string in format 'YYYY-MM-DDTh:m:s'
        """
        return datetime.strptime(string, "%Y-%m-%dT%H:%M:%S")

    @staticmethod
    def get_datetime_from_date_hour(date: str, hour: int) -> datetime:
        """
        Returns a datetime object from a date and an hour
        """
        if hour == 24:
            hour = 0
        return datetime.strptime(f"{date}T{hour}", "%Y-%m-%dT%H")

    @staticmethod
    def get_date_hour_from_string(string: str, zero_to_twenty_four=False):
        """
        Returns a tuple (date, hour) from an hour-string
        """
        date_t = datetime.strptime(string, "%Y-%m-%dT%H:%M:%S")
        hour = date_t.strftime("%H")
        if hour == "00" and zero_to_twenty_four:
            hour = "24"
            date_t -= timedelta(days=1)
        date = date_t.strftime("%Y-%m-%d")
        return date, int(hour)

    @staticmethod
    def get_date_string_from_ts(ts: datetime) -> str:
        """Returns the string of a given date as 'YYYY-MM-DD'"""
        return datetime.strftime(ts, "%Y-%m-%d")

    @staticmethod
    def get_datetime_string_from_ts(ts: datetime) -> str:
        """Returns the string of a given date as 'YYYY-MM-DDTh:m'"""
        return datetime.strftime(ts, "%Y-%m-%dT%H:%M")

    @staticmethod
    def get_datetimesec_string_from_ts(ts: datetime) -> str:
        """Returns the string of a given date as 'YYYY-MM-DDTh:m:s'"""
        return datetime.strftime(ts, "%Y-%m-%dT%H:%M:%S")

    @staticmethod
    def get_next_hour_datetime_string(string: str) -> str:
        """
        Returns the hour following the given hour, as a string
        """
        date_t = datetime.strptime(string, "%Y-%m-%dT%H:%M:%S")
        return (date_t + timedelta(hours=1)).isoformat()

    @staticmethod
    def get_next_hour_datetimesec_string(string: str) -> str:
        """
        Returns the hour following the given hour, as a string
        """
        date_t = datetime.strptime(string, "%Y-%m-%dT%H:%M:%S")
        return (date_t + timedelta(hours=1)).isoformat()

    @staticmethod
    def get_next_hour(ts: datetime) -> datetime:
        """
        Returns the hour following the given hour
        """
        return ts + timedelta(hours=1)

    @staticmethod
    def get_previous_hour_datetime_string(string: str) -> str:
        """
        Returns the hour preceding the given hour, as a string
        """
        date_t = datetime.strptime(string, "%Y-%m-%dT%H:%M:%S")
        return (date_t - timedelta(hours=1)).isoformat()

    @staticmethod
    def get_previous_hour_datetimesec_string(string: str) -> str:
        """
        Returns the hour preceding the given hour, as a string
        """
        date_t = datetime.strptime(string, "%Y-%m-%dT%H:%M:%S.%f")
        return (date_t - timedelta(hours=1)).isoformat()

    @staticmethod
    def get_previous_hour(ts: datetime) -> datetime:
        """
        Returns the hour preceding the given hour
        """
        return ts - timedelta(hours=1)

    @staticmethod
    def get_date_string_from_ts_string(ts: str) -> str:
        """Returns the date in format 'YYYY-MM-DD' from a datetime string"""
        return ts[0:10]

    @staticmethod
    def get_hour_from_ts(ts: datetime) -> float:
        """Returns the hours (in number) of the given time slot"""
        return float(ts.hour + ts.minute / 60)

    @staticmethod
    def add_time_to_ts(ts: datetime, weeks=0, days=0, minutes=0, seconds=0) -> datetime:
        """Adds time to a datetime"""
        return ts + timedelta(days=7 * weeks + days, minutes=minutes, seconds=seconds)

    @staticmethod
    def add_time_to_date_string(string: str, weeks=0, days=0, minutes=0, seconds=0) -> str:
        """Adds time to a date string"""
        date = datetime.strptime(string, "%Y-%m-%d").date()
        return (date + timedelta(days=7 * weeks + days, minutes=minutes, seconds=seconds)).strftime("%Y-%m-%d")

    @staticmethod
    def add_time_to_datetime_string(string: str, weeks=0, days=0, minutes=0, seconds=0) -> str:
        """Adds time to a datetime"""
        datetime_object = datetime.strptime(string, "%Y-%m-%dT%H:%M:%S")
        return (datetime_object + timedelta(days=7 * weeks + days, minutes=minutes, seconds=seconds)).strftime("%Y-%m-%dT%H:%M:%S")

    @staticmethod
    def add_time_to_datetimesec_string(string: str, weeks=0, days=0, hours=0, minutes=0, seconds=0) -> str:
        """Adds time to a datetime"""
        datetime_object = datetime.strptime(string, "%Y-%m-%dT%H:%M:%S")
        return (datetime_object + timedelta(weeks=weeks, days=7 * weeks + days, hours=hours, minutes=minutes, seconds=seconds)).strftime("%Y-%m-%dT%H:%M:%S")

    @staticmethod
    def get_week_from_ts(ts: datetime) -> int:
        """Returns the integer value of the week for the given time slot"""
        return ts.isocalendar()[1]

    @staticmethod
    def get_week_from_date_string(string: str) -> int:
        """Returns the integer value of the week for the given string"""
        date_object = datetime.strptime(string, "%Y-%m-%d").date()
        return date_object.isocalendar()[1]

    @staticmethod
    def get_week_from_datetime_string(string: str) -> int:
        """Returns the integer value of the week for the given string"""
        datetime_object = datetime.strptime(string, "%Y-%m-%dT%H:%M:%S")
        return datetime_object.isocalendar()[1]

    @staticmethod
    def get_week_from_datetimesec_string(string: str) -> int:
        """Returns the integer value of the week for the given string"""
        datetime_object = datetime.strptime(string, "%Y-%m-%dT%H:%M:%S")
        return datetime_object.isocalendar()[1]

    @staticmethod
    def get_weekday_from_ts(ts: datetime) -> int:
        """Returns the number of the weekday from a ts"""
        return ts.isocalendar()[2]

    @staticmethod
    def get_weekday_from_date_string(string: str) -> int:
        """Returns the number of the weekday from a date string in format 'YYYY-MM-DD'"""
        date = datetime.strptime(string, '%Y-%m-%d').date()
        return date.isocalendar()[2]

    @staticmethod
    def get_weekday_from_datetime_string(string: str) -> int:
        """Returns the number of the weekday from a date string in format 'YYYY-MM-DDTh:m'"""
        datetime_obj = datetime.strptime(string, '%Y-%m-%dT%H:%M:%S')
        return datetime_obj.isocalendar()[2]

    @staticmethod
    def get_weekday_from_datetimesec_string(string: str) -> int:
        """Returns the number of the weekday from a date string in format 'YYYY-MM-DDT:h:m:s'"""
        datetime_obj = datetime.strptime(string, '%Y-%m-%dT%H:%M:%S')
        return datetime_obj.isocalendar()[2]

    @staticmethod
    def get_hour_from_datetime_string(string: str) -> float:
        """Returns the integer value of the hour (in number) from ts string in format 'YYYY-MM-DDTh:m'"""
        datetime_obj = datetime.strptime(string, '%Y-%m-%dT%H:%M:%S')
        return datetime_obj.hour

    @staticmethod
    def get_hour_from_datetimesec_string(string: str) -> float:
        """Returns the integer value of the hour (in number) from ts string in format 'YYYY-MM-DDTh:m:s'"""
        datetime_obj = datetime.strptime(string, '%Y-%m-%dT%H:%M:%S')
        return datetime_obj.hour



"""

"""
# Full imports
import json
import logging
import os

# Partial imports
from datetime import datetime, timedelta
from urllib.parse import urlparse, urljoin


# Imports from modules
from cornflow_client import CornFlow, CornFlowApiError


# TODO: convert everything to an object that encapsulates everything
#  to make it clear and avoid all the arguments.
# class DagApp(object):
#
#     def __init__(self, dag_name, secrets):
#         self.name = dag_name
#         self.secrets = secrets
#
#     def solve(self, data, config):
#         raise NotImplemented()
#
#     def test_cases(self):
#         raise NotImplemented()
#
#     pass


default_args = {
    "owner": "baobab",
    "depends_on_past": False,
    "start_date": datetime(2020, 2, 1),
    "email": [""],
    "email_on_failure": False,
    "email_on_retry": False,
    "retry_delay": timedelta(minutes=1),
    "schedule_interval": None,
}


def get_schemas_from_file(_dir, dag_name):
    with open(os.path.join(_dir, dag_name + "_input.json"), "r") as f:
        instance = json.load(f)
    with open(os.path.join(_dir, dag_name + "_output.json"), "r") as f:
        solution = json.load(f)
    return instance, solution


def get_requirements(path):
    """
    Read requirements.txt from a project and return a list of packages.

    :param path: The path of the project
    :return: A list of required packages
    """
    req_path = f"{path}/requirements.txt"

    try:
        with open(req_path, "r") as file:
            req_list = file.read().splitlines()
    except:
        print("no requirement file in this path")
        return []

    return req_list


def connect_to_cornflow(secrets):
    """
    Create a connection to cornflow and log in with airflow admin user.

    :return: A logged and connected Cornflow class instance
    """
    # This secret comes from airflow configuration
    print("Getting connection information from ENV VAR=CF_URI")
    uri = secrets.get_conn_uri("CF_URI")
    conn = urlparse(uri)
    scheme = conn.scheme
    if scheme == "cornflow":
        scheme = "http"
    url = f"{scheme}://{conn.hostname}"
    if conn.port:
        url = f"{url}:{conn.port}"
    if conn.path:
        url = urljoin(url, conn.path)
    airflow_user = CornFlow(url=url)
    airflow_user.login(username=conn.username, pwd=conn.password)
    return airflow_user


def try_to_save_error(client, exec_id, state=-1):
    """
    Attempt at saving that the execution failed
    """
    try:
        client.put_api_for_id("dag/", id=exec_id, payload=dict(state=state))
    except Exception as e:
        print(f"An exception trying to register the failed status: {e}")


def try_to_save_airflow_log(client, exec_id, ti, base_log_folder):
    log_file = os.path.join(
        base_log_folder,
        f"{ti.dag_id}",
        f"{ti.task_id}",
        f"{ti.run_id}".replace("manual__", "").replace("scheduled__", ""),
        f"{ti.try_number}.log"
    )
    if os.path.exists(log_file):
        with open(log_file, 'r') as fd:
            log_file_txt = fd.read()
        try:
            client.put_api_for_id("dag/", id=exec_id, payload=dict(log_text=log_file_txt))
        except Exception as e:
            print(f"An exception occurred while trying to register airflow log: {e}")


def try_to_write_solution(client, exec_id, payload):
    """
    Tries to write the payload into cornflow
    If it fails tries to write again that it failed.
    If it fails at least once: it raises an exception
    """
    execution_payload = dict(**payload)
    execution_payload.pop("inst_id")
    execution_payload.pop("inst_checks")
    try:
        client.write_solution(execution_id=exec_id, **execution_payload)
    except CornFlowApiError:
        try_to_save_error(client, exec_id, -6)
        # attempt to update the execution with a failed status.
        raise AirflowDagException("The writing of the solution failed")

    if payload.get("inst_checks") is not None:
        checks_payload = dict()
        checks_payload["checks"] = payload["inst_checks"]
        try:
            client.write_instance_checks(
                instance_id=payload["inst_id"], **checks_payload
            )
        except CornFlowApiError:
            try_to_save_error(client, exec_id, -6)
            raise AirflowDagException("The writing of the instance checks failed")


def get_schema(dag_name):
    _file = os.path.join(os.path.dirname(__file__), f"{dag_name}_output.json")
    with open(_file, "r") as f:
        schema = json.load(f)
    return schema


def cf_solve_app(app, secrets, **kwargs):
    if kwargs["dag_run"].conf.get("checks_only"):
        return cf_check(app.check, app.name, secrets, **kwargs)
    else:
        return cf_solve(app.solve, app.name, secrets, **kwargs)


def cf_solve(fun, dag_name, secrets, **kwargs):
    """
    Connect to cornflow, ask for data, solve the problem and write the solution in cornflow

    :param fun: The function to use to solve the problem.
    :param dag_name: the name of the dag, to later search the output schema
    :param kwargs: other kwargs passed to the dag task.
    :return:
    """
    ti = kwargs["ti"]
    base_log_folder = kwargs["conf"].get("logging", "base_log_folder")
    config = dict()
    try:
        client = connect_to_cornflow(secrets)
        exec_id = kwargs["dag_run"].conf["exec_id"]
        execution_data = client.get_data(exec_id)
        execution_status = client.update_status(exec_id, {"status": 0})
        data = execution_data["data"]
        solution_data = execution_data["solution_data"]
        config = execution_data["config"]
        inst_id = execution_data["id"]

        solution, sol_checks, inst_checks, log, log_json = fun(
            data, config, solution_data
        )

        # We connect again to cornflow in case that more than 24 hours
        # have passed from the first time we connect
        client = connect_to_cornflow(secrets)

        payload = dict(
            state=1,
            log_json=log_json,
            log_text=log,
            solution_schema=dag_name,
            inst_checks=inst_checks,
            inst_id=inst_id,
        )
        if not solution:
            # No solution found: we just send everything to cornflow.
            if config.get("msg", True):
                print("No solution found: we save what we have.")
            try_to_write_solution(client, exec_id, payload)
            return "Solution was not saved"
        # There is a solution:
        # we first need to validate the schema.
        # If it's not: we change the status to Invalid
        # and take out the server validation of the schema\
        payload["data"] = solution

        if config.get("msg", True):
            print("A solution was found: we will first validate it")

        if sol_checks is not None:
            payload["checks"] = sol_checks

        try_to_write_solution(client, exec_id, payload)

        # The validation went correctly: can save the solution without problem
        return "Solution saved"

    except NoSolverException as e:
        if config.get("msg", True):
            print("No solver found !")
        try_to_save_error(client, exec_id, -1)
        client.update_status(exec_id, {"status": -1})
        try_to_save_airflow_log(client, exec_id, ti, base_log_folder)
        raise AirflowDagException(e)
    except Exception as e:
        if config.get("msg", True):
            print("Some unknown error happened")
        try_to_save_error(client, exec_id, -1)
        client.update_status(exec_id, {"status": -1})
        try_to_save_airflow_log(client, exec_id, ti, base_log_folder)
        raise AirflowDagException("There was an error during the solving")


def cf_check(fun, dag_name, secrets, **kwargs):
    """
    Connect to cornflow, ask for data, check the solution data and write the checks in cornflow
    :param fun: The function to use to check the data
    :param secrets: Environment variables
    :param kwargs: other kwargs passed to the dag task.
    :return:
    """
    ti = kwargs["ti"]
    base_log_folder = kwargs["conf"].get("logging", "base_log_folder")
    config = dict()
    try:
        client = connect_to_cornflow(secrets)
        exec_id = kwargs["dag_run"].conf["exec_id"]
        execution_data = client.get_data(exec_id)
        config = execution_data["config"]

        instance_data = execution_data["data"]
        inst_id = execution_data["id"]
        solution_data = execution_data["solution_data"]

        inst_checks, sol_checks, log_json = fun(instance_data, solution_data)

        if config.get("checks_only"):
            payload = dict(
                state=1,
                log_json=log_json,
                log_text="Data checked.",
                solution_schema=dag_name,
                inst_checks=inst_checks,
                inst_id=inst_id,
            )
        else:
            payload = dict(
                inst_checks=inst_checks,
                inst_id=inst_id,
                state=1,
            )

        if sol_checks is not None:
            payload["checks"] = sol_checks

        try_to_write_solution(client, exec_id, payload)

        case_id = kwargs["dag_run"].conf.get("case_id")
        if case_id is not None:
            checks_payload = dict(checks=payload["inst_checks"])
            if sol_checks is not None:
                checks_payload["solution_checks"] = sol_checks
            try:
                client.write_case_checks(case_id=case_id, **checks_payload)
            except CornFlowApiError:
                try_to_save_error(client, exec_id, -6)
                if config.get("msg", True):
                    print("An error occurred while trying to save the checks")
                try_to_save_airflow_log(client, exec_id, ti, base_log_folder)
                raise AirflowDagException("The writing of the case checks failed")

        # # The validation went correctly: can save the solution without problem
        return "Checks saved"

    except Exception as e:
        if config.get("msg", True):
            print("Some unknown error happened")
        try_to_save_error(client, exec_id, -1)
        client.update_status(exec_id, {"status": -1})
        try_to_save_airflow_log(client, exec_id, ti, base_log_folder)
        raise AirflowDagException(
            "There was an error during the verification of the data"
        )


class NoSolverException(Exception):
    pass


class AirflowDagException(Exception):
    pass



"""

"""
import json
import os
import pickle
from datetime import datetime, timedelta
from unittest import TestCase
from cornflow_client.core import InstanceSolutionCore

from cornflow_client import InstanceCore

from cornflow_client.schema.tools import get_empty_schema


# Constants
path_to_tests_dir = os.path.dirname(os.path.abspath(__file__))

# Helper functions
def _load_file(_file):
    with open(_file) as f:
        temp = json.load(f)
    return temp


def _get_file(relative_path):
    return os.path.join(path_to_tests_dir, relative_path)


class TestSimpleApplicationDag(TestCase):
    def setUp(self):
        self.class_to_use = SimpleInstance
        self.instance_filename = _get_file("../data/gc_input.json")
        self.instance_data = _load_file(self.instance_filename)
        self.schema = get_empty_schema()

    def tearDown(self):
        pass

    def test_instance_from_dict(self):
        instance = self.class_to_use.from_dict(self.instance_data)
        self.assertEqual(self.instance_data, instance.data)
        return instance

    def test_instance_to_dict(self):
        instance = self.test_instance_from_dict()
        self.assertEqual(self.instance_data, instance.to_dict())

    def test_instance_from_json(self):
        instance = self.class_to_use.from_json(self.instance_filename)
        self.assertEqual(self.instance_data, instance.data)
        return instance

    def test_instance_to_json(self):
        instance = self.test_instance_from_json()
        instance.to_json(_get_file("../data/gc_input_export.json"))
        read_data = _load_file(_get_file("../data/gc_input.json"))
        self.assertEqual(self.instance_data, read_data)
        os.remove(_get_file("../data/gc_input_export.json"))

    def test_instance_schema(self):
        instance = self.test_instance_from_dict()
        self.assertEqual(self.schema, instance.schema)

    def test_schema_validation(self):
        instance = self.test_instance_from_dict()
        self.assertEqual([], instance.check_schema())

    def test_schema_generation(self):
        instance = self.test_instance_from_dict()
        schema = _load_file(_get_file("../data/graph_coloring_input.json"))
        schema["properties"]["pairs"].pop("description")
        self.assertEqual(schema, instance.generate_schema())

    def test_to_excel(self):
        instance = self.test_instance_from_dict()
        instance.to_excel(_get_file("../data/gc_input_export.xlsx"))
        instance_2 = self.class_to_use.from_excel(
            _get_file("../data/gc_input_export.xlsx")
        )
        self.assertEqual(instance.data, instance_2.data)
        os.remove(_get_file("../data/gc_input_export.xlsx"))

    def test_dict_to_int_or_float(self):
        # Test with simple input
        input_dict = {'a': '4', 'b': '7.8'}
        expected_output = {'a': 4, 'b': 7.8}
        self.assertEqual(
            InstanceSolutionCore.dict_to_int_or_float(input_dict),
            expected_output
        )

        # Test with nested dictionary and list
        input_dict = {
            'a': '4',
            'b': {'c': '7', 'd': ['8.7', '9']}
        }
        expected_output = {
            'a': 4,
            'b': {'c': 7, 'd': [8.7, 9]}
        }
        self.assertEqual(
            InstanceSolutionCore.dict_to_int_or_float(input_dict),
            expected_output
        )
        input_dict = {
            'a': '4',
            'b': {'c': '7', 'd': 'not_a_number'}
        }
        expected_output = {
            'a': 4,
            'b': {'c': 7, 'd': 'not_a_number'}
        }
        self.assertEqual(
            InstanceSolutionCore.dict_to_int_or_float(input_dict),
            expected_output
        )

    def test_from_element_or_list_to_dict(self):
        # Test with list input
        input_list = [{'index': 4, 'value': 5}, {'index': 7, 'value': 8}]
        expected_output = {
            4: {'index': 4, 'value': 5},
            7: {'index': 7, 'value': 8}
        }
        self.assertEqual(
            InstanceSolutionCore.from_element_or_list_to_dict(input_list),
            expected_output
        )

        # Test with single element input
        input_dict = {'index': 4, 'value': 5}
        expected_output = {4: {'index': 4, 'value': 5}}
        self.assertEqual(
            InstanceSolutionCore.from_element_or_list_to_dict(input_dict),
            expected_output
        )

    def test_get_date_from_string(self):
        date_string = '2022-01-01'
        expected_output = datetime.strptime(date_string, '%Y-%m-%d')
        self.assertEqual(
            InstanceSolutionCore.get_date_from_string(date_string),
            expected_output
        )

    def test_get_datetime_from_string(self):
        datetime_string = '2022-01-01T12:34'
        expected_output = datetime.strptime(datetime_string, '%Y-%m-%dT%H:%M')
        self.assertEqual(
            InstanceSolutionCore.get_datetime_from_string(datetime_string),
            expected_output
        )

    def test_get_datetimesec_from_string(self):
        string = "2022-01-01T00:00:00"
        expected_result = datetime(2022, 1, 1, 0, 0, 0)
        self.assertEqual(InstanceSolutionCore.get_datetimesec_from_string(string), expected_result)

    def test_get_datetime_from_date_hour(self):
        date = "2022-01-01"
        hour = 0
        expected_result = datetime(2022, 1, 1, 0)
        self.assertEqual(InstanceSolutionCore.get_datetime_from_date_hour(date, hour), expected_result)

    def test_get_date_hour_from_string_with_zero_to_twenty_four(self):
        # Test with zero_to_twenty_four set to True
        string = "2021-01-01T00:00:00"
        date, hour = InstanceSolutionCore.get_date_hour_from_string(string, True)
        self.assertEqual(date, "2020-12-31")
        self.assertEqual(hour, 24)

    def test_get_date_hour_from_string_without_zero_to_twenty_four(self):
        # Test with zero_to_twenty_four set to False
        string = "2021-01-01T00:00:00"
        date, hour = InstanceSolutionCore.get_date_hour_from_string(string, False)
        self.assertEqual(date, "2021-01-01")
        self.assertEqual(hour, 0)

    def test_get_date_hour_from_string_with_noon_time(self):
        # Test with a string representing noon time
        string = "2021-01-01T12:00:00"
        date, hour = InstanceSolutionCore.get_date_hour_from_string(string, False)
        self.assertEqual(date, "2021-01-01")
        self.assertEqual(hour, 12)

    def test_get_date_string_from_ts(self):
        ts = datetime(2022, 1, 1, 0, 0, 0)
        expected_result = "2022-01-01"
        self.assertEqual(InstanceSolutionCore.get_date_string_from_ts(ts), expected_result)

    def test_get_datetime_string_from_ts(self):
        ts = datetime(2022, 1, 1, 0, 0, 0)
        expected_result = "2022-01-01T00:00"
        self.assertEqual(InstanceSolutionCore.get_datetime_string_from_ts(ts), expected_result)

    def test_get_datetimesec_string_from_ts(self):
        ts = datetime(2022, 1, 1, 0, 0, 0)
        expected_result = "2022-01-01T00:00:00"
        self.assertEqual(InstanceSolutionCore.get_datetimesec_string_from_ts(ts), expected_result)

    def test_get_next_hour_datetime_string(self):
        string = "2022-01-01T00:00:00"
        expected_result = "2022-01-01T01:00:00"
        self.assertEqual(InstanceSolutionCore.get_next_hour_datetime_string(string), expected_result)

    def test_get_next_hour_datetimesec_string(self):
        string = "2022-01-01T00:00:00"
        expected_result = "2022-01-01T01:00:00"
        self.assertEqual(InstanceSolutionCore.get_next_hour_datetimesec_string(string), expected_result)

    def test_get_next_hour(self):
        ts = datetime(2022, 1, 1, 12)
        expected = datetime(2022, 1, 1, 13)
        self.assertEqual(InstanceSolutionCore.get_next_hour(ts), expected)

    def test_get_previous_hour_datetime_string(self):
        input_string = "2022-10-11T15:45:30"
        expected_output = "2022-10-11T14:45:30"
        self.assertEqual(InstanceSolutionCore.get_previous_hour_datetime_string(input_string), expected_output)

    def test_get_previous_hour_datetimesec_string(self):
        input_string = "2022-10-11T15:45:30.123456"
        expected_output = "2022-10-11T14:45:30.123456"
        self.assertEqual(InstanceSolutionCore.get_previous_hour_datetimesec_string(input_string), expected_output)

    def test_get_previous_hour(self):
        input_ts = datetime(2022, 10, 11, 15, 45, 30)
        expected_output = datetime(2022, 10, 11, 14, 45, 30)
        self.assertEqual(InstanceSolutionCore.get_previous_hour(input_ts), expected_output)

    def test_get_date_string_from_ts_string(self):
        input_ts = "2022-10-11T15:45:30"
        expected_output = "2022-10-11"
        self.assertEqual(InstanceSolutionCore.get_date_string_from_ts_string(input_ts), expected_output)

    def test_get_hour_from_ts(self):
        input_ts = datetime(2022, 10, 11, 15, 45, 30)
        expected_output = 15.75
        self.assertEqual(InstanceSolutionCore.get_hour_from_ts(input_ts), expected_output)

    def test_add_time_to_ts(self):
        input_ts = datetime(2022, 10, 11, 15, 45, 30)
        expected_output = datetime(2022, 10, 18, 15, 45, 30)
        self.assertEqual(InstanceSolutionCore.add_time_to_ts(input_ts, days=7), expected_output)

    def test_add_time_to_date_string(self):
        instance = self.class_to_use.from_dict(self.instance_data)
        initial_date = "2022-10-11"
        expected_date = (datetime.strptime(initial_date, "%Y-%m-%d").date()
                        + timedelta(days=7)).strftime("%Y-%m-%d")
        result = instance.add_time_to_date_string(initial_date, days=7)
        self.assertEqual(result, expected_date)

    def test_add_time_to_datetime_string(self):
        initial_datetime = "2022-01-01T12:00:00"
        expected_datetime = (datetime.strptime(initial_datetime, "%Y-%m-%dT%H:%M:%S")
                            + timedelta(weeks=2, days=3, minutes=45, seconds=30)).strftime("%Y-%m-%dT%H:%M:%S")
        result = InstanceSolutionCore.add_time_to_datetime_string(initial_datetime, weeks=2, days=3, minutes=45, seconds=30)
        self.assertEqual(result, expected_datetime)

    def test_add_time_to_datetimesec_string(self):
        string = "2022-01-01T12:00:00"
        result = InstanceSolutionCore.add_time_to_datetimesec_string(string, days=1)
        expected = "2022-01-02T12:00:00"
        self.assertEqual(result, expected)

    def test_get_week_from_ts(self):
        ts = datetime(2022, 1, 1)
        result = InstanceSolutionCore.get_week_from_ts(ts)
        expected = 52
        self.assertEqual(result, expected)

    def test_get_week_from_date_string(self):
        string = "2022-01-01"
        result = InstanceSolutionCore.get_week_from_date_string(string)
        expected = 52
        self.assertEqual(result, expected)

    def test_get_week_from_datetime_string(self):
        string = "2022-01-01T12:00:00"
        result = InstanceSolutionCore.get_week_from_datetime_string(string)
        expected = 52
        self.assertEqual(result, expected)

    def test_get_week_from_datetimesec_string(self):
        string = "2022-01-01T12:00:00"
        result = InstanceSolutionCore.get_week_from_datetimesec_string(string)
        expected = 52
        self.assertEqual(result, expected)

    def test_get_weekday_from_ts(self):
        ts = datetime(2022, 1, 1)
        result = InstanceSolutionCore.get_weekday_from_ts(ts)
        expected = 6
        self.assertEqual(result, expected)

    def test_get_weekday_from_date_string(self):
        string = "2022-01-01"
        result = InstanceSolutionCore.get_weekday_from_date_string(string)
        expected = 6
        self.assertEqual(result, expected)

    def test_get_weekday_from_datetime_string(self):
        string = "2022-01-01T12:00:00"
        result = InstanceSolutionCore.get_weekday_from_datetime_string(string)
        expected = 6
        self.assertEqual(result, expected)

    def test_get_weekday_from_datetimesec_string(self):
        string = "2022-01-01T12:00:00"
        result = InstanceSolutionCore.get_weekday_from_datetimesec_string(string)
        expected = 6
        self.assertEqual(result, expected)

    def test_get_hour_from_datetime_string(self):
        datetime_string = "2023-01-01T12:00:00"
        expected_hour = 12
        result = InstanceSolutionCore.get_hour_from_datetime_string(datetime_string)
        self.assertEqual(result, expected_hour)

    def test_get_hour_from_datetimesec_string(self):
        datetime_string = "2023-01-01T12:00:00"
        expected_hour = 12
        result = InstanceSolutionCore.get_hour_from_datetimesec_string(datetime_string)
        self.assertEqual(result, expected_hour)



######################################################
class TestCustomInstanceDag(TestSimpleApplicationDag):
    def setUp(self):
        super().setUp()
        self.class_to_use = Instance
        self.schema = _load_file(_get_file("../data/graph_coloring_input.json"))

    def test_instance_from_dict(self):
        """Here as to be not equal as the data has been processed on the from_dict method"""
        instance = self.class_to_use.from_dict(self.instance_data)
        self.assertNotEqual(self.instance_data, instance.data)
        return instance

    def test_instance_from_json(self):
        """Here as to be not equal as the data has been processed on the from_dict method"""
        instance = self.class_to_use.from_json(self.instance_filename)
        self.assertNotEqual(self.instance_data, instance.data)
        return instance


class SimpleInstance(InstanceCore):
    schema = get_empty_schema()
    schema_checks = get_empty_schema()


class Instance(InstanceCore):
    schema = _load_file(_get_file("../data/graph_coloring_input.json"))
    schema_checks = get_empty_schema()

    @classmethod
    def from_dict(cls, data):
        tables = ["pairs"]

        data_p = {el: {(v["n1"], v["n2"]): v for v in data[el]} for el in tables}

        return cls(data_p)

    def to_dict(self):
        tables = ["pairs"]

        data_p = {el: self.data[el].values_l() for el in tables}
        return pickle.loads(pickle.dumps(data_p, -1))



"""
Constants values used in schemas functions.
"""
import pulp as pl
from ortools.sat.python import cp_model


STRING_TYPE = "String"
BOOLEAN_TYPE = "Boolean"
INTEGER_TYPE = "Integer"
FLOAT_TYPE = "Float"
BASIC_TYPES = [STRING_TYPE, BOOLEAN_TYPE, INTEGER_TYPE, FLOAT_TYPE]

JSON_TYPES = {
    "string": STRING_TYPE,
    "number": FLOAT_TYPE,
    "integer": INTEGER_TYPE,
    "null": None,
    "boolean": BOOLEAN_TYPE,
}

DATASCHEMA = "DataSchema"

INSTANCE_SCHEMA = "instance"
SOLUTION_SCHEMA = "solution"
CONFIG_SCHEMA = "config"
INSTANCE_CHECKS_SCHEMA = "instance_checks"
SOLUTION_CHECKS_SCHEMA = "solution_checks"

# why it stopped
STATUS_NOT_SOLVED = 0
STATUS_OPTIMAL = 1
STATUS_INFEASIBLE = -1
STATUS_UNBOUNDED = -2
STATUS_UNDEFINED = -3
STATUS_FEASIBLE = 2
STATUS_MEMORY_LIMIT = 3
STATUS_NODE_LIMIT = 4
STATUS_TIME_LIMIT = 5
STATUS_LICENSING_PROBLEM = -5
STATUS_QUEUED = -7

# Associated string
STATUS_CONV = {
    STATUS_OPTIMAL: "Optimal",
    STATUS_TIME_LIMIT: "Time limit",
    STATUS_INFEASIBLE: "Infeasible",
    STATUS_UNDEFINED: "Unknown",
    STATUS_NOT_SOLVED: "Not solved",
    STATUS_UNBOUNDED: "Unbounded",
    STATUS_FEASIBLE: "Feasible",
    STATUS_MEMORY_LIMIT: "Memory limit",
    STATUS_NODE_LIMIT: "Node limit",
    STATUS_LICENSING_PROBLEM: "Licensing problem",
    STATUS_QUEUED: "Queued"
}

# is there a solution?
SOLUTION_STATUS_INFEASIBLE = 0
SOLUTION_STATUS_FEASIBLE = 2

PYOMO_STOP_MAPPING = {
    "unbounded": STATUS_UNBOUNDED,
    "infeasible": STATUS_INFEASIBLE,
    "invalidProblem": STATUS_NOT_SOLVED,
    "solverFailure": STATUS_NOT_SOLVED,
    "internalSolverError": STATUS_NOT_SOLVED,
    "error": STATUS_NOT_SOLVED,
    "userInterrupt": STATUS_NOT_SOLVED,
    "resourceInterrupt": STATUS_NOT_SOLVED,
    "licensingProblem": STATUS_LICENSING_PROBLEM,
    "maxTimeLimit": STATUS_TIME_LIMIT,
    "maxIterations": STATUS_TIME_LIMIT,
    "maxEvaluations": STATUS_TIME_LIMIT,
    "globallyOptimal": STATUS_OPTIMAL,
    "locallyOptimal": STATUS_OPTIMAL,
    "optimal": STATUS_OPTIMAL,
    "minFunctionValue": STATUS_UNDEFINED,
    "minStepLength": STATUS_UNDEFINED,
    "other": STATUS_UNDEFINED,
    "intermediateNonInteger": STATUS_TIME_LIMIT,
    "feasible": STATUS_TIME_LIMIT,
    "noSolution": STATUS_TIME_LIMIT,
}

PYOMO_STATUS_MAPPING = {
    "ok": SOLUTION_STATUS_FEASIBLE,
    "warning": SOLUTION_STATUS_INFEASIBLE,
    "error": SOLUTION_STATUS_INFEASIBLE,
    "aborted": SOLUTION_STATUS_INFEASIBLE,
    "unknown": SOLUTION_STATUS_INFEASIBLE,
}

ORTOOLS_STATUS_MAPPING = {
    cp_model.OPTIMAL: STATUS_OPTIMAL,
    cp_model.FEASIBLE: STATUS_FEASIBLE,
    cp_model.INFEASIBLE: STATUS_INFEASIBLE,
    cp_model.UNKNOWN: STATUS_UNDEFINED,
    cp_model.MODEL_INVALID: STATUS_UNDEFINED,
}

PULP_STATUS_MAPPING = {
    pl.LpStatusInfeasible: STATUS_INFEASIBLE,
    pl.LpStatusNotSolved: STATUS_NOT_SOLVED,
    pl.LpStatusOptimal: STATUS_OPTIMAL,
    pl.LpStatusUnbounded: STATUS_UNBOUNDED,
    pl.LpStatusUndefined: STATUS_UNDEFINED,
}

PARAMETER_SOLVER_TRANSLATING_MAPPING = {
    ('abs_gap', 'pyomo', 'cbc') : 'allow' ,
    ('rel_gap', 'pyomo', 'cbc'): 'ratio' ,
    ('cutoff', 'pyomo', 'cbc') : 'cuts' ,
    ('bar_iter_limit', 'pyomo', 'cbc') : 'barr' ,
    ('best_obj_stop', 'pyomo', 'cbc') : 'primalT' ,
    ('iteration_limit', 'pyomo', 'cbc') : 'maxIt' ,
    ('solution_limit', 'pyomo', 'cbc') : 'maxSaved' ,
    ('time_limit', 'pyomo', 'cbc') : 'sec' ,
    ('pump_passes', 'pyomo', 'cbc') : 'pumpC' ,
    ('heuristics', 'pyomo', 'cbc') : 'heur' ,

    ('is_mip', 'pulp', 'cbc') : 'mip' ,
    ('abs_gap', 'pulp', 'cbc') : 'gapAbs' ,
    ('rel_gap', 'pulp', 'cbc') : 'gapRel' ,
    ('cutoff', 'pulp', 'cbc') : 'cuts' ,
    ('time_limit', 'pulp', 'cbc') : 'timeLimit' ,
    ('threads', 'pulp', 'cbc') : 'threads' ,
    ('presolve', 'pulp', 'cbc') : 'presolve' ,
    

    ('abs_gap', 'pyomo', 'gurobi') : 'MIPGapAbs' ,
    ('rel_gap', 'pyomo', 'gurobi') : 'MIPGap' ,
    ('time_limit', 'pyomo', 'gurobi') : 'TimeLimit' ,
    ('presolve', 'pyomo', 'gurobi') : 'Presolve' ,
    ('iteration_limit', 'pyomo', 'gurobi') : 'IterationLimit' , 
    ('bar_iter_limit', 'pyomo', 'gurobi') : 'BarIterLimit' ,
    ('best_obj_stop', 'pyomo', 'gurobi') : 'BestObjStop' ,
    ('cutoff', 'pyomo', 'gurobi') : 'Cutoff' ,
    ('mem_limit', 'pyomo', 'gurobi') : 'MemLimit' ,
    ('solution_limit', 'pyomo', 'gurobi') : 'SolutionLimit' ,
    ('branch_dir', 'pyomo', 'gurobi') : 'BranchDir' ,
    ('pump_passes', 'pyomo', 'gurobi') : 'PumpPasses' ,
    ('heuristics', 'pyomo', 'gurobi') : 'Heuristics' ,
    ('threads', 'pyomo', 'gurobi') : 'threads' ,
    
    ('is_mip', 'pulp', 'gurobi') : 'mip' ,
    ('abs_gap', 'pulp', 'gurobi') : 'gapAbs' ,
    ('rel_gap', 'pulp', 'gurobi') : 'gapRel' ,
    ('time_limit', 'pulp', 'gurobi') : 'timeLimit' ,
    ('presolve', 'pulp', 'gurobi') : 'presolve' ,
    ('feasibility_tol', 'pulp', 'gurobi') : 'FeasibilityTol' ,
    ('iteration_limit', 'pulp', 'gurobi') : 'IterationLimit' ,


    ('abs_gap', 'pyomo', 'scip') : 'limits/absgap ' ,
    ('rel_gap', 'pyomo', 'scip') : 'limits/gap' ,
    ('time_limit', 'pyomo', 'scip') : 'limits/time' ,
    ('threads', 'pyomo', 'scip') : 'threads' ,
    ('bar_tol', 'pyomo', 'scip') : 'numerics/barrierconvtol' ,
    ('cutoff_breaker', 'pyomo', 'scip') : 'heuristics/shiftandpropagate/cutoffbreaker' ,
    ('iteration_limit', 'pyomo', 'scip') : 'lp/iterlim' ,
    ('mem_limit', 'pyomo', 'scip') : 'limits/memory' ,
    ('solution_limit', 'pyomo', 'scip') : 'limits/maxsol' ,
    ('pump_passes', 'pyomo', 'scip') : 'heuristics/feaspump/maxdepth' ,
    ('presolve', 'pyomo', 'scip') : 'presolve' ,
    ('nlp_tol', 'pyomo', 'scip') : 'heuristics/subnlp/opttol' ,
    ('cutoff', 'pyomo', 'scip') : 'heuristics/subnlp/setcutoff' ,
    ('nlp_iteration_limit', 'pyomo', 'scip') : 'heuristics/subnlp/itermin' ,

    ('is_mip', 'pulp', 'scip') : 'mip' ,
    ('abs_gap', 'pulp', 'scip') : 'gapAbs' ,
    ('rel_gap', 'pulp', 'scip') : 'gapRel' ,
    ('time_limit', 'pulp', 'scip') : 'timeLimit' ,
    ('threads', 'pulp', 'scip') : 'threads' ,
    ('max_nodes', 'pulp', 'scip') : 'maxNodes' ,
    ('iteration_limit', 'pulp', 'scip') : 'lp/iterlim' ,
    ('mem_limit', 'pulp', 'scip') : 'limits/memory' ,
    ('cutoff_breaker', 'pulp', 'scip') : 'heuristics/shiftandpropagate/cutoffbreaker' ,
    ('presolve', 'pulp', 'scip') : 'presolve' ,
    ('solution_limit', 'pulp', 'scip') : 'limits/maxsol' ,
    ('nlp_iteration_limit', 'pulp', 'scip') : 'heuristics/subnlp/itermin' ,


    ('rel_gap', 'pyomo', 'highs') : 'mip_rel_gap' ,
    ('abs_gap', 'pyomo', 'highs') : 'mip_abs_gap' ,
    ('time_limit', 'pyomo', 'highs') : 'time_limit' ,
    ('presolve', 'pyomo', 'highs') : 'presolve' ,
    ('parallel', 'pyomo', 'highs') : 'parallel' ,
    ('crossover', 'pyomo', 'highs') : 'run_crossover' ,
    ('heuristics', 'pyomo', 'highs') : 'mip_heuristic_effort' ,

    ('is_mip', 'pulp', 'highs') : 'mip' ,
    ('abs_gap', 'pulp', 'highs') : 'gapAbs' ,
    ('rel_gap', 'pulp', 'highs') : 'gapRel' ,
    ('time_limit', 'pulp', 'highs') : 'timeLimit' ,
    ('presolve', 'pulp', 'highs') : 'presolve' ,
    ('threads', 'pulp', 'highs') : 'threads' ,
    ('heuristics', 'pulp', 'highs') : 'mip_heuristic_effort' ,

}




class AirflowError(Exception):
    status_code = 400
    log_txt = "Airflow error"

    def __init__(self, error=None, status_code=None, payload=None, log_txt=None):
        Exception.__init__(self, error)
        if error is not None:
            self.error = error
        if status_code is not None:
            self.status_code = status_code
        self.payload = payload
        if log_txt is not None:
            self.log_txt = log_txt
        else:
            self.log_txt = self.error

    def to_dict(self):
        rv = dict(self.payload or ())
        rv["error"] = self.error
        return rv


class NoSolverException(Exception):
    pass


class BadConfiguration(Exception):
    pass


class BadInstance(Exception):
    pass


class BadSolution(Exception):
    pass



"""

"""
# Partial imports
from abc import ABC, abstractmethod

# Imports from internal modules
from .instance_solution import InstanceSolutionCore


class InstanceCore(InstanceSolutionCore, ABC):
    """
    The instance template.
    """

    # TODO: make abstractmethod
    def check(self, *args, **kwargs) -> dict:
        """
        Method that checks if there are inconsistencies in the data of the instance and if the problem is feasible

        :return: An dictionary containing the inconsistencies found and indicating if the problem is infeasible
        """
        return dict()

    @property
    @abstractmethod
    def schema_checks(self) -> dict:
        """
        A dictionary representation of the json-schema for the dictionary returned by the method Instance.check()
        """
        raise NotImplementedError()



from .raw_cornflow_client import RawCornFlow, CornFlowApiError


class CornFlow:
    def __init__(self, url, token=None):
        self.raw = RawCornFlow(url, token)

        self.sign_up = self.expect_status(self.raw.sign_up, 201)
        self.create_instance = self.expect_status(self.raw.create_instance, 201)
        self.create_case = self.expect_status(self.raw.create_case, 201)
        self.create_instance_file = self.expect_status(self.raw.create_instance_file, 201)
        self.create_execution = self.expect_status(self.raw.create_execution, 201)
        self.relaunch_execution = self.expect_status(self.raw.relaunch_execution, 201)
        self.create_execution_data_check = self.expect_status(self.raw.create_execution_data_check, 201)
        self.create_instance_data_check = self.expect_status(self.raw.create_instance_data_check, 201)
        self.create_case_data_check = self.expect_status(self.raw.create_case_data_check, 201)
        self.get_data = self.expect_status(self.raw.get_data, 200)
        self.write_solution = self.expect_status(self.raw.write_solution, 200)
        self.write_instance_checks = self.expect_status(self.raw.write_instance_checks, 200)
        self.write_case_checks = self.expect_status(self.raw.write_case_checks, 200)
        self.stop_execution = self.expect_status(self.raw.stop_execution, 200)
        self.manual_execution = self.expect_status(self.raw.manual_execution, 201)
        self.get_results = self.expect_status(self.raw.get_results, 200)
        self.get_status = self.expect_status(self.raw.get_status, 200)
        self.update_status = self.expect_status(self.raw.update_status, 200)
        self.get_log = self.expect_status(self.raw.get_log, 200)
        self.get_solution = self.expect_status(self.raw.get_solution, 200)
        self.get_all_instances = self.expect_status(self.raw.get_all_instances, 200)
        self.get_all_cases = self.expect_status(self.raw.get_all_cases, 200)
        self.get_all_executions = self.expect_status(self.raw.get_all_executions, 200)
        self.get_all_users = self.expect_status(self.raw.get_all_users, 200)
        self.get_one_user = self.expect_status(self.raw.get_one_user, 200)
        self.get_one_instance = self.expect_status(self.raw.get_one_instance, 200)
        self.get_one_instance_data = self.expect_status(self.raw.get_one_instance_data, 200)
        self.get_one_case = self.expect_status(self.raw.get_one_case, 200)
        self.get_one_case_data = self.expect_status(self.raw.get_one_case_data, 200)
        self.put_one_case = self.expect_status(self.raw.put_one_case, 200)
        self.put_one_instance = self.expect_status(self.raw.put_one_instance, 200)
        self.put_one_execution = self.expect_status(self.raw.put_one_execution, 200)
        self.patch_one_case = self.expect_status(self.raw.patch_one_case, 200)
        self.delete_one_instance = self.expect_status(self.raw.delete_one_instance, 200)
        self.delete_one_case = self.expect_status(self.raw.delete_one_case, 200)
        self.delete_one_execution = self.expect_status(self.raw.delete_one_execution, 200)
        self.get_schema = self.expect_status(self.raw.get_schema, 200)
        self.get_all_schemas = self.expect_status(self.raw.get_all_schemas, 200)
        self.get_deployed_dags = self.expect_status(self.raw.get_deployed_dags, 200)
        self.create_deployed_dag = self.expect_status(self.raw.create_deployed_dag, 201)
        self.put_deployed_dag = self.expect_status(self.raw.put_deployed_dag, 200)

    @property
    def url(self):
        """ Gets the url of the server """
        return self.raw.url

    @url.setter
    def url(self, url):
        """ Sets the url of the server """
        self.raw.url = url

    @property
    def token(self):
        """ Gets the token """
        return self.raw.token

    @token.setter
    def token(self, token):
        """ Sets the token """
        self.raw.token = token

    @staticmethod
    def expect_status(func, expected_status=None):
        """
        Gets the response of the call
        and raise an exception if the status of the response is not the expected
        """
        def decorator(*args, **kwargs):
            response = func(*args, **kwargs)
            if expected_status is not None and response.status_code != expected_status:
                raise CornFlowApiError(
                    f"Expected a code {expected_status}, got a {response.status_code} error instead: {response.text}"
                )
            return response.json()
        return decorator

    def is_alive(self):
        """
        Asks the server if it's alive
        """
        response = self.raw.is_alive()
        if response.status_code == 200:
            return response.json()
        raise CornFlowApiError(
            f"Connection failed with status code: {response.status_code}: {response.text}"
        )

    def login(self, username, pwd, encoding=None):
        """
        Log-in to the server.

        :param str username: username
        :param str pwd: password
        :param str encoding: the type of encoding used in the call. Defaults to 'br'

        :return: a dictionary with a token inside
        """
        response = self.raw.login(username, pwd, encoding=encoding)
        if response.status_code != 200:
            raise CornFlowApiError(
                f"Login failed with status code: {response.status_code}: {response.text}"
            )
        return response.json()
